package squircle;






/**
 * Vect2 is a class for storing and manipulating vectors and positions in 2D.<br/>
 */


public class Vect2 
{
	

	/** The x component of the vector. @related y @example Vect2_x */
	public float x;
	
	/** The y component of the vector. @related x @example Vect2_y */
	public float y;
	
	public float z;

	
	// for internal calculations //
	/** The Constant TWO_PI. */
	public static final float TWO_PI = (float) ( Math.PI * 2f );
	
//	public float[] arr= {x,y};
	
	
	//////////////////
	// CONSTRUCTORS //
	//////////////////
	
	
	/**
	 * Instantiates a new vect2.
	 */
	public Vect2(){ setZero(); }

	/**
	 * Instantiates a new vect2.
	 *
	 * @param x the x
	 * @param y the y
	 */
	public Vect2( float x, float y ){ set( x, y ); }

	/**
	 * Instantiates a new vect2.
	 *
	 * @param vector to copy into new vector (<code>Vect2</code>)
	 */
	public Vect2( Vect2 vector ){ set( vector ); }
	
	/**
	 * Alternative constructor to create a unit vector at a specific angle.
	 *
	 * @param angle the angle
	 */
	public Vect2( float angle )throws Exception
	{
		x = (float) Math.cos( angle );
		y = (float) Math.sin( angle );
	}
	
	
	
	////////////////
	// PROPERTIES //
	////////////////
	
	
	/**
	 * Returns a boolean that is true if the vector is zero.
	 * In other words, it returns ( x==0 && y==0 ).
	 *
	 * @return true if vector is zero (<code>boolean</code>)
	 * @shortdesc Returns <code>true</code> if the vector is zero.
	 * @related isAlmostZero ( )
	 * @example Vect2_is_zero
	 */
	public boolean isZero() throws Exception{
		return (x == 0f && y == 0f);
	}
	
	
	/**
	 * Returns a boolean that is true if the vector is almost zero
	 * ( +/- 0.001f ). This is useful when easing towards zero to
	 * check if the target has been reached.
	 *
	 * @return true if vector is almost zero (<code>boolean</code>)
	 * @shortdesc Returns <code>true</code> if the vector is almost zero.
	 * @related isZero ( )
	 * @example Vect2_is_almost_zero
	 */
	public boolean isAlmostZero()throws Exception{
		return( Math.abs( x ) < 0.1f && Math.abs( y ) < 0.1f );
	}
	
	
	/**
	 * Returns a boolean that is true if the vector is longer than the
	 * input magnitude. This function is about ten times faster than
	 * a regular distance check, because it avoids using square root and
	 * division.
	 *
	 * @param magnitude to compare against (<code>int</code>)
	 * @return true if the vector is longer than the given magnitude (<code>boolean</code>)
	 * @shortdesc Returns <code>true</code> if the vector is longer than a specific magnitude.
	 * @related isLessThan ( )
	 * @related isZero ( )
	 * @related isAlmostZero ( )
	 * @example Vect2_is_greater_than
	 */
	public boolean isGreaterThan( float magnitude ) throws Exception{
		return( squareMagnitude() > magnitude * magnitude );
	}
	
	/**
	 * Checks if is greater than.
	 *
	 * @param magnitude the magnitude
	 * @return true, if is greater than
	 */
	public boolean isGreaterThan( int magnitude ) throws Exception{
		return( squareMagnitude() > magnitude * magnitude );
	}
	
	
	/**
	 * Returns a boolean that is true if the vector is shorter than the
	 * input magnitude. This function is about ten times faster than
	 * a regular distance check, because it avoids using square root and
	 * division.
	 *
	 * @param magnitude to compare against (<code>int</code>)
	 * @return true if the vector is shorter than the given magnitude (<code>boolean</code>)
	 * @shortdesc Returns <code>true</code> if the vector is shorter than a specific magnitude.
	 * @related isGreaterThan ( )
	 * @related isZero ( )
	 * @related isAlmostZero ( )
	 * @example Vect2_is_less_than
	 */
	public boolean isLessThan( float magnitude ) throws Exception{
		return( squareMagnitude() < magnitude * magnitude );
	}
	
	/**
	 * Checks if is less than.
	 *
	 * @param magnitude the magnitude
	 * @return true, if is less than
	 */
	public boolean isLessThan( int magnitude ) throws Exception{
		return( squareMagnitude() < magnitude * magnitude );
	}
	
	
	/**
	 * Returns a boolean that is true if the vector is equals an input vector.
	 *
	 * @param vector to compare against (<code>Vect2</code>)
	 * @return true if the given vector is equal this vector (<code>boolean</code>)
	 * @shortdesc Returns <code>true</code> if the vector is equals another vector
	 * @example Vect2_equals
	 */
	public boolean equals( Vect2 vector ){
		if( vector == null ) return false;
		return( this.x == vector.x && this.y == vector.y );
	}
	
	
	
	
	
	////////////////
	// BASIC MATH //
	////////////////
	
	
	/**
	 * Multiplies the vector by an input value and stores the result.
	 *
	 * @param multiplier to scale this vector (<code>float</code>)
	 * @shortdesc Multiplies the vector by an input value and stores the result.
	 * @related scaled ( )
	 * @related divide ( )
	 * @related add ( )
	 * @related subtract ( )
	 * @example Vect2_scale
	 */
	public void scale( float multiplier )throws Exception{
		x *= multiplier;
		y *= multiplier;
	}
	
	/**
	 * Scale.
	 *
	 * @param multX the mult x
	 * @param multY the mult y
	 */
	public void scale( float multX, float multY ){
		x *= multX;
		y *= multY;
	}
	
	/**
	 * Scale.
	 *
	 * @param multVect the mult vect
	 */
	public void scale( Vect2 multVect )throws Exception{
		x *= multVect.x;
		y *= multVect.y;
	}
	
	
	/**
	 * Multiplies the vector by an input value and returns the result without changing the original.
	 *
	 * @param multiplier to scale this vector (<code>float</code>)
	 * @return new vector containing the result of the multiplication (<code>Vect2</code>)
	 * @shortdesc Multiplies the vector by an input value and returns the result.
	 * @related scale ( )
	 * @related divided ( )
	 * @related added ( )
	 * @related subtracted ( )
	 * @example Vect2_scaled
	 */
	public Vect2 scaled( float multiplier ){
		Vect2 result = this.copy();
		result.x *= multiplier;
		result.y *= multiplier;
		return result;
	}
	
	/**
	 * Scaled.
	 *
	 * @param multX the mult x
	 * @param multY the mult y
	 * @return the vect2
	 */
	public Vect2 scaled( float multX, float multY )throws Exception{
		Vect2 result = this.copy();
		result.x *= multX;
		result.y *= multY;
		return result;
	}
	
	/**
	 * Scaled.
	 *
	 * @param multVector the mult vector
	 * @return the vect2
	 */
	public Vect2 scaled( Vect2 multVector )throws Exception{
		Vect2 result = this.copy();
		result.x *= multVector.x;
		result.y *= multVector.y;
		return result;
	}

	
	/**
	 * Divides the vector by a divider and stores the result.
	 *
	 * @param divider to divide this vector (<code>float</code>)
	 * @shortdesc Divides the vector by a divider and stores the result.
	 * @related divided ( )
	 * @related scale ( )
	 * @related add ( )
	 * @related subtract ( )
	 * @example Vect2_divide
	 */
	public void divide( float divider ) throws Exception{
		if( divider != 0 ){
			scale( 1 / divider );
		}
	}
	
	
	/**
	 * Divides the vector by a divider and returns the result without changing the original.
	 *
	 * @param divider to divide this vector (<code>float</code>)
	 * @return new vector containing the result of the division (<code>Vect2</code>)
	 * @shortdesc Divides the vector by a divider and returns the result.
	 * @related divide ( )
	 * @related scaled ( )
	 * @related added ( )
	 * @related subtracted ( )
	 * @example Vect2_divided
	 */
	public Vect2 divided( float divider ) throws Exception{
		if( divider != 0 ){
			Vect2 result = this.copy();
			result.scale( 1 / divider );
			return result;
		}
		return new Vect2( 0, 0 );
	}
	
    /**
     * Returns the dot product of two vectors.  Generally, this can b interpreted as the
     * angle between them.
     * @param a compare to this vector
     * @return the dot product
     */
  public float dot( Vect2 a)throws Exception{
    return (x*a.x) + (y*a.y);
  }

	/**
	 * Returns the euclidean distance between this and another vector.
	 *
	 * @param vector2 containing a position (<code>Vect2</code>)
	 * @return distance between the two vectors (<code>float</code>)
	 * @shortdesc Returns the  distance between to positions.
	 * @example Vect2_distance
	 */

	public float distance(  Vect2 vector2 ){
		float dx = x - vector2.x;
		float dy = y - vector2.y;
		return (float) Math.sqrt( dx * dx + dy * dy );
	}
	/**
	 * Returns the square distance between this and another vector.
	 *
	 * @param vector2 containing a position (<code>Vect2</code>)
	 * @return square-distance between the two vectors (<code>float</code>)
	 * @shortdesc Returns the square distance between to positions.
	 * @example Vect2_distance
	 */

	public float squareDistance(  Vect2 vector2 ) {
		
		return distance(vector2)*distance(vector2);
		
	}

	

	/**
	 * Adds an input vector to the vector and stores the result.
	 *
	 * @param x the x
	 * @param y the y
	 * @shortdesc Adds an input vector to the vector and stores the result.
	 * @related added ( )
	 * @related subtract ( )
	 * @related scale ( )
	 * @related divide ( )
	 * @example Vect2_add
	 */
	public void add( float x, float y ){
		this.x += x;
		this.y += y;
	}
	
	/**
	 * Adds the.
	 *
	 * @param vector the vector
	 */
	public void add( Vect2 vector ) throws Exception{
		this.x += vector.x;
		this.y += vector.y;
	}
	
	
	/**
	 * Returns the result of an input vector added to the vector without changing the originals.
	 *
	 * @param x the x
	 * @param y the y
	 * @return new vector containing the result of the addition (<code>Vect2</code>)
	 * @shortdesc Adds an input vector to the vector and returns the result.
	 * @related add ( )
	 * @related subtracted ( )
	 * @related scaled ( )
	 * @related divided ( )
	 * @example Vect2_added
	 */
	public Vect2 added( float x, float y )throws Exception{
		Vect2 result = this.copy();
		result.x += x;
		result.y += y;
		return result;
	}
	
	/**
	 * Added.
	 *
	 * @param vector the vector
	 * @return the vect2
	 */
	public Vect2 added( Vect2 vector ) throws Exception{
		Vect2 result = this.copy();
		result.x += vector.x;
		result.y += vector.y;
		return result;
	}
	
	
	/**
	 * Subtracts an input from the vector and stores the result.<br />
	 *
	 * @param x the x
	 * @param y the y
	 * @shortdesc Subtracts an input from the vector and stores the result.
	 * @related subtracted ( )
	 * @related add ( )
	 * @related scale ( )
	 * @related divide ( )
	 * @example Vect2_subtract
	 */
	public void subtract( float x, float y )throws Exception{
		this.x -= x;
		this.y -= y;
	}
	
	/**
	 * Subtract.
	 *
	 * @param vector the vector
	 */
	public void subtract( Vect2 vector )throws Exception{
		this.x -= vector.x;
		this.y -= vector.y;
	}
	
	/**
	 * Lerped.
	 *
	 * @param vect2 
	 * @param t the t
	 * @return new vect2
	 */
	public Vect2 lerped( Vect2 vector,float t )throws Exception{
		return new Vect2((vector.x-this.x)*t+this.x,(vector.y-this.y)*t+this.y);
		
	}

    	  
	
	/**
	 * Subtracts and input vector from the vector and returns the result without changing the originals.
	 *
	 * @param x the x
	 * @param y the y
	 * @return new vector containing the result of the subtraction (<code>Vect2</code>)
	 * @shortdesc Subtracts and input vector from the vector and returns the result.
	 * @related subtract ( )
	 * @related added ( )
	 * @related scaled ( )
	 * @related divided ( )
	 * @example Vect2_subtracted
	 */
	public Vect2 subtracted( float x, float y )throws Exception{
		Vect2 result = this.copy();
		result.x -= x;
		result.y -= y;
		return result;
	}
	
	/**
	 * Subtracted.
	 *
	 * @param vector the vector
	 * @return the vect2
	 */
	public Vect2 subtracted( Vect2 vector )throws Exception{
		Vect2 result = this.copy();
		result.x -= vector.x;
		result.y -= vector.y;
		return result;
	}
	
	
	
	///////////////
	// MAGNITUDE //
	///////////////
	
	
	/**
	 * Returns the magnitude of the vector, also referred to as the length.
	 * 
	 * @return magnitude of this vector (<code>float</code>)
	 * @related setMagnitude ( )
	 * @example Vect2_magnitude
	 */
	public float magnitude()throws Exception
	{
		return (float) Math.sqrt( ( x * x ) + ( y * y ) );
	}
	
	
	/**
	 * Returns the square magnitude (square length) of the vector. <br>
	 * <br>
	 * The square root function which is used when calling magnitude() is horrendously slow,
	 * so try to avoid calculating vector lengths whenever you can. A common problem in computer
	 * graphics is to find the shortest vector in a list, in this case you only need to calculate
	 * the square magnitude (x*x + y*y) for each of them and find the smallest value from that
	 * (since the vector with the shortest length will also have the smallest squared length).
	 *
	 * @return square magnitude of this vector (<code>float</code>)
	 * @shortdesc Returns the square magnitude of the vector.
	 * @related magnitude ( )
	 * @related setMagnitude ( )
	 * @example Vect2_square_magnitude
	 */
	public float squareMagnitude()throws Exception
	{
		return x * x + y * y;
	}
	
	
	/**
	 * Normalizes the vector. Same as calling <code>setMagnitude( 1 )</code>.
	 * 
	 * @shortdesc Normalizes the vector.
	 * @related normalized ( )
	 * @related setMagnitude ( )
	 * @example Vect2_normalize
	 */
	public void normalize() throws Exception{
		setMagnitude( 1 );
	}
	
	
	/**
	 * Returns a copy that has been normalized.
	 *
	 * @return new normalized copy of the vector (<code>Vect2</code>)
	 * @shortdesc Returns a copy that has been normalized.
	 * @related normalize ( )
	 * @related setMagnitude ( )
	 * @example Vect2_normalized
	 */
	public Vect2 normalized() throws Exception{
		Vect2 normalized = new Vect2( this.x, this.y );
		normalized.normalize();
		return normalized;
	}
	
	
	/**
	 * Sets the magnitude (length) of the vector.
	 *
	 * @param magnitude (<code>float</code>)
	 * @related normalize ( )
	 * @example Vect2_set_magnitude
	 */
	public void setMagnitude( int magnitude ) throws Exception{
		if( !isZero() ) {
			float mult = ( (float) magnitude ) / magnitude();
			scale(mult);
		}
	}
	
	/**
	 * Sets the magnitude.
	 *
	 * @param magnitude the new magnitude
	 */
	public void setMagnitude( float magnitude )throws Exception {
		if( !this.isZero() ){
			float mult = magnitude / magnitude();
			this.scale( mult );
		}
	}
	
	
	/**
	 * Clips the magnitude of the vector. In other words, if the
	 * magnitude exceeds the input value, it is set to that value.
	 *
	 * @param maximum magnitude (<code>float</code>)
	 * @shortdesc Clips the magnitude of the vector.
	 * @related clipped ( )
	 * @related setMagnitude ( )
	 * @example Vect2_clip
	 */
	public void clip( float maximum ) throws Exception{
		if( magnitude() > maximum ) setMagnitude( maximum );
	}
	
	
	/**
	 * Returns a copy of the vector that has been clipped.
	 *
	 * @param maximum magnitude (<code>float</code>)
	 * @return the vect2
	 * @related clip ( )
	 * @related setMagnitude ( )
	 * @example Vect2_clipped
	 */
	public Vect2 clipped( float maximum )throws Exception {
		Vect2 result = this.copy();
		if( result.magnitude() > maximum ) result.setMagnitude( maximum );
		return result;
	}
	
	
	
	
	//////////////
	// ROTATION //
	//////////////
	
	
	/**
	 * Returns the angle of the vector in radians. All rotations are based on
	 * the unit circle where angle 0 and PI*2 is 3 o'clock.
	 *
	 * @return angle of this vector in radians (<code>float</code>)
	 * @shortdesc Returns the angle of the vector.
	 * @example Vect2_angle
	 */
	public float angle()throws Exception
	{
		float a = (float) Math.atan2( y, x );
		if( a < 0 ) {
			a += TWO_PI;
			if( a == TWO_PI ) a = 0;
			return a;
		} else {
			return a;
		}
	}
	
	
	/**
	 * Rotates the vector clockwise by a delta angle
	 * (turning left from the vectors forward direction).
	 *
	 * @param deltaAngle in radians (<code>float</code>)
	 * @shortdesc Rotates the vector clockwise by a a delta angle.
	 * @related rotated ( )
	 * @related setRotation ( )
	 * @example Vect2_rotate
	 */
	public void rotate( float deltaAngle )throws Exception{
		float x = this.x;
		float y = this.y;
		this.x = (float) ( x * Math.cos( deltaAngle ) - y * Math.sin( deltaAngle ) );
		this.y = (float) ( x * Math.sin( deltaAngle ) + y * Math.cos( deltaAngle ) );
	}
	
	
	/**
	 * Returns a copy that is rotated clockwise from the vector by a specified angle
	 * (turning left from the vectors forward direction).
	 *
	 * @param deltaAngle the delta angle
	 * @return a copy that has been rotated (<code>Vect2</code>)
	 * @shortdesc Returns copy that has been rotated clockwise by a delta angle.
	 * @related rotate ( )
	 * @related setRotation ( )
	 * @example Vect2_rotated
	 */
	public Vect2 rotated( float deltaAngle )throws Exception{
		Vect2 result = this.copy();
		result.rotate( deltaAngle );
		return result;
	}
	
	/**
	 * Rotated around.
	 *returns a copy of a Vect2 clockwise rotated around a vector
	 * @param position the position
	 * @param deltaAngle the delta angle
	 * @return the vect2
	 */
	public Vect2 rotatedAround(Vect2 position, float deltaAngle)throws Exception{
		Vect2 t = this.copy();
		t.subtract(position);
		t.rotate(deltaAngle);
		t.add(position);
		
		return t;
		
	}
	
	/**
	 * Rotate around.
	 * rotates a vect2 clockwise around a vector
	 * @param position the position
	 * @param deltaAngle the delta angle
	 */
	public void rotateAround(Vect2 position, float deltaAngle)throws Exception{
		subtract(position);
		rotate(deltaAngle);
		add(position);
	}
	

	/**
	 * Sets the rotation of the vector. All rotations are based on
	 * the unit circle where angle 0 and PI*2 is 3 o'clock. It is slightly
	 * faster to call rotate() by a delta angle.
	 *
	 * @param angle in radians (<code>float</code>)
	 * @shortdesc Sets the rotation of the vector.
	 * @related rotate ( )
	 * @example Vect2_set_rotation
	 */
	public void setRotation( float angle )throws Exception
	{
		float len = this.magnitude();
		set( len, 0 );
		rotate( angle );
	}
	
	
	/**
	 * Rotates the vector 90 degrees right. The magnitude is left unchanged.
	 * This function is much faster than rotate() or setRotation().
	 * 
	 * @shortdesc Rotates the vector 90 degrees right.
	 * @related rotateLeft ( )
	 * @related rotatedLeft ( )
	 * @related rotatedRight ( )
	 * @related rotated ( )
	 * @related rotate ( )
	 * @related setRotation ( )
	 * @example Vect2_rotate_right
	 */
	public void rotateRight()throws Exception{
		float newX = y;
		y = x;
		x = -newX;
	}
	
	
	/**
	 * Returns a copy that has been rotated 90 degrees clockwise from the vector's forward angle.
	 * The magnitude is unchanged.
	 * This function is much faster than rotate() or setRotation().
	 *
	 * @return a copy that has been rotated 90 degrees clockwise (<code>Vect2</code>)
	 * @shortdesc Returns a copy that has been rotated 90 degrees clockwise.
	 * @related rotatedLeft ( )
	 * @related rotateLeft ( )
	 * @related rotateRight ( )
	 * @related rotated ( )
	 * @related rotate ( )
	 * @related setRotation ( )
	 * @example Vect2_rotated_right
	 */
	public Vect2 rotatedRight()throws Exception
	{
		Vect2 v = new Vect2( -y, x );
        return v;
    }
	
	
	/**
	 * Rotates the vector 90 degrees left. The magnitude is left unchanged.
	 * This function is much faster than rotate() or setRotation().
	 * 
	 * @shortdesc Rotates the vector 90 degrees left.
	 * @related rotateRight ( )
	 * @related rotatedLeft ( )
	 * @related rotatedRight ( )
	 * @related rotated ( )
	 * @related rotate ( )
	 * @related setRotation ( )
	 * @example Vect2_rotate_left
	 */
	public void rotateLeft()throws Exception{
		float newX = y;
		y = -x;
		x = newX;
	}
	
	
	/**
	 * Returns a copy that has been rotated 90 degrees counter clockwise from the vector's forward angle.
	 * The magnitude is unchanged.
	 * This function is much faster than rotate() or setRotation().
	 *
	 * @return a copy that has been rotated 90 degrees counter clockwise (<code>Vect2</code>)
	 * @shortdesc Returns a copy that has been rotated 90 degrees counter clockwise.
	 * @related rotatedRight ( )
	 * @related rotateLeft ( )
	 * @related rotateRight ( )
	 * @related rotated ( )
	 * @related rotate ( )
	 * @related setRotation ( )
	 * @example Vect2_rotated_left
	 */
	public Vect2 rotatedLeft()throws Exception
	{
		Vect2 v =  new Vect2( y, -x );
        return v;
    }
	
	
	/**
	 * Flips the vector 180 degrees from it's forward angle.
	 * 
	 * @shortdesc Flips the vector 180 degrees.
	 * @example Vect2_flip
	 */
	public void flip()throws Exception{
		x *= -1;
		y *= -1;
	}
	
	
	/**
	 * Returns a copy that is flipped 180 degrees from the vector's forward angle.
	 *
	 * @return new vector that is flipped 180 degrees from the vector (<code>Vect2</code>)
	 * @shortdesc Returns a copy that has been flipped.
	 * @related flip ( )
	 * @related rotatedRight ( )
	 * @related rotatedLeft ( )
	 * @example Vect2_flipped
	 */
	public Vect2 flipped() throws Exception{
		return new Vect2( -x, -y );
	}
	/**
	 * Returns a copy that is flipped 180 degrees from the vector's forward angle.
	 *
	 * @return new vector that is flipped 180 degrees from the vector (<code>Vect2</code>)
	 * @shortdesc Returns a copy that has been flipped.
	 * @related flip ( )
	 * @related rotatedRight ( )
	 * @related rotatedLeft ( )
	 * @example Vect2_flipped
	 */
	public Vect2 flippedX() throws Exception{
		return new Vect2( -x, y );
	}
	/**
	 * Returns a copy that is flipped 180 degrees from the vector's forward angle.
	 *
	 * @return new vector that is flipped 180 degrees from the vector (<code>Vect2</code>)
	 * @shortdesc Returns a copy that has been flipped.
	 * @related flip ( )
	 * @related rotatedRight ( )
	 * @related rotatedLeft ( )
	 * @example Vect2_flipped
	 */
	public Vect2 flippedY() throws Exception{
		return new Vect2( x, -y );
	}
	
	
	
	
	
	//////////////////
	// STATIC TOOLS //
	//////////////////
	
	
	/**
	 * Returns the euclidean distance between to positions stored in
	 * two vectors.
	 *
	 * @param vector1 containing a position (<code>Vect2</code>)
	 * @param vector2 containing a position (<code>Vect2</code>)
	 * @return distance between the two vectors (<code>float</code>)
	 * @shortdesc Returns the  distance between to positions.
	 * @example Vect2_distance
	 */
	public static float distance( Vect2 vector1, Vect2 vector2 )throws Exception {
		float dx = vector1.x - vector2.x;
		float dy = vector1.y - vector2.y;
		return (float) Math.sqrt( dx * dx + dy * dy );
	}
	
	
	/**
	 * Returns the dot product of this vector and another.
	 * The dot product value is in fact the cosine of the angle
	 * between the two input vectors, multiplied by the lengths
	 * of those vectors. So, you can easily calculate the cosine
	 * of the angle by either, making sure that your two vectors
	 * are both of length 1, or dividing the dot product by the
	 * lengths.
	 * cos( theta ) = dot( v1, v2 ) / ( magnitude( v1 ) * magnitude( v2 ) )
	 * Values range from 1 to -1. If the two input vectors are
	 * pointing in the same direction, then the return value will
	 * be 1. If the two input vectors are pointing in opposite
	 * directions, then the return value will be -1. If the two
	 * input vectors are at right angles, then the return value
	 * will be 0. So, in effect, it is telling you how similar
	 * the two vectors are.
	 *
	 * @param vector1 (<code>Vect2</code>)
	 * @param vector2 (<code>Vect2</code>)
	 * @return dot product (<code>float</code>)
	 * @shortdesc Returns the dot product of this vector and another.
	 * @example Vect2_dot
	 */
	public static float dot( Vect2 vector1, Vect2 vector2 )throws Exception
	{
        float dot = vector1.x * vector2.x + vector1.y * vector2.y;
        return dot;
    }
	
	
	/**
	 * Returns the linear interpolation between two vectors. In some vector
	 * classes this is referred to as blend().
	 *
	 * @param begin the vector when interpolator is 0 (<code>Vect2</code>)
	 * @param end the vector when interpolator is 1 (<code>Vect2</code>)
	 * @param interpolator to lerp between the two vectors (<code>float</code>)
	 * @return interpolated vector (<code>Vect2</code>)
	 * @shortdesc Returns the linear interpolation between two vectors.
	 * @example Vect2_lerp
	 */
	public static Vect2 lerp( Vect2 begin, Vect2 end, float interpolator )throws Exception
	{
		Vect2 v = new Vect2();
		v.x = ( end.x - begin.x ) * interpolator + begin.x;
		v.y = ( end.y - begin.y ) * interpolator + begin.y;
        return v;
    }
	
	
	/**
	 * Returns the angle between the forward angles of two vectors.<br>
	 * <br>
	 * The result is always positive. If you also need to know the rotational direction
	 * (clockwise or counter-clockwise) you need to calculate the angle of each vector
	 * and work it out from there.
	 *
	 * @param vector1 the vector1
	 * @param vector2 the vector2
	 * @return angle in radians, or NaN if vectors are parallel (<code>float</code>)
	 * @shortdesc Returns the angle between two vectors.
	 * @example Vect2_angle_between
	 */
	public static float angleBetween( Vect2 vector1, Vect2 vector2 ) throws Exception{
		float theta = dot( vector1.normalized(), vector2.normalized() );
		return (float) Math.acos( theta );
	}
	
	
	/**
	 * Returns the angle between the forward angles of two unit vectors.
	 * Use angleBetween() for vectors that are not normalized. This function is only included
	 * to occasionally save a couple of internal normalize calls.<br>
	 * <br>
	 * The result is always positive. If you also need to know the rotational direction
	 * (clockwise or counter-clockwise) you need to calculate the angle of each vector
	 * and work it out from there.
	 *
	 * @param vector1 the vector1
	 * @param vector2 the vector2
	 * @return angle in radians, or NaN if vectors are parallel (<code>float</code>)
	 * @shortdesc Returns the angle between two unit vectors.
	 * @related angleBetween ( )
	 * @example Vect2_angle_between_units
	 */
	public static float angleBetweenUnits( Vect2 vector1, Vect2 vector2 ) throws Exception{
		return (float) Math.acos( dot( vector1, vector2 ) );
	}
	
	
	
	/////////////
	// HELPERS //
	/////////////
	
	
	/**
	 * Sets the vector by copying (no reference is kept).
	 *
	 * @param x the x
	 * @param y the y
	 * @shortdesc Sets the vector by copying.
	 * @related setZero ( )
	 * @example Vect2_set
	 */
	public void set( float x, float y ){
		this.x = x;
		this.y = y;
	}
	
	/**
	 * Sets the.
	 *
	 * @param v the v
	 */
	public void set( Vect2 v ){
		x = v.x;
		y = v.y;
	}

	
	/**
	 * Sets the vector to zero (<code>0,0</code>).
	 * 
	 * @related set ( )
	 * @example Vect2_set_zero
	 */
	public void setZero(){
		x = 0f;
		y = 0f;
	}
	
	
	/**
	 * Returns a copy of the vector. Use this when you want to avoid creating
	 * references between vectors; instead of doing <code>v1 = v2</code>, do
	 * <code>v1 = v2.copy</code>
	 *
	 * @return new copy of this vector (<code>Vect2</code>)
	 * @shortdesc Returns a copy of the vector.
	 * @example Vect2_copy
	 */
	public Vect2 copy(){
		return new Vect2( this.x, this.y );
	}
	
	public boolean isValid(){
	    
	    if( !Float.isNaN(this.x) && !Float.isNaN(this.y))
	        return true;
	    
	    
        return false;
	    
	}
	
	/**
	 * Returns a string containing the (<code>x,y</code>) components of this vector.
	 *
	 * @return the string
	 * @returns string describing this vector
	 * @example Vect2_to_string
	 */
	public String toString() {
		return "Vect2( " + x + ", " + y + " )";
	}
	
}
